#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>

#define N_ELEM 10000000

// First test: use the same thing as the C version.

typedef struct {
  int x, y, z;
} int_vec3;

typedef struct {
  float x, y, z;
} float_vec3;

#define VEC3_SUM_MACRO(_c, _a, _b) {                 \
    (_c).x = (_a).x + (_b).x;			\
    (_c).y = (_a).y + (_b).y;			\
    (_c).z = (_a).z + (_b).z;			\
  }

void test_int1() {
  int_vec3 a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = rand();
    a[i].y = rand();
    a[i].z = rand();
    b[i].x = rand();
    b[i].y = rand();
    b[i].z = rand();
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    VEC3_SUM_MACRO(c[i], b[i], a[i]);
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with integers in C equivalent: %ld\n", t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

void test_float1() {
  float_vec3 a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = ((float)rand())/RAND_MAX;
    a[i].y = ((float)rand())/RAND_MAX;
    a[i].z = ((float)rand())/RAND_MAX;
    b[i].x = ((float)rand())/RAND_MAX;
    b[i].y = ((float)rand())/RAND_MAX;
    b[i].z = ((float)rand())/RAND_MAX;
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    VEC3_SUM_MACRO(c[i], b[i], a[i]);
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with floats in C equivalent: %ld\n", t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

// Second test: use C structs with template function instead of macro.

template<typename T>
void vec3_sum(T &c, T const &a, T const &b) {
  c.x = a.x + b.x;
  c.y = a.y + b.y;
  c.z = a.z + b.z;
}

void test_int2() {
  int_vec3 a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = rand();
    a[i].y = rand();
    a[i].z = rand();
    b[i].x = rand();
    b[i].y = rand();
    b[i].z = rand();
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    vec3_sum(c[i], b[i], a[i]);
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with integers using template function: %ld\n",
         t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

void test_float2() {
  float_vec3 a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = ((float)rand())/RAND_MAX;
    a[i].y = ((float)rand())/RAND_MAX;
    a[i].z = ((float)rand())/RAND_MAX;
    b[i].x = ((float)rand())/RAND_MAX;
    b[i].y = ((float)rand())/RAND_MAX;
    b[i].z = ((float)rand())/RAND_MAX;
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    vec3_sum(c[i], b[i], a[i]);
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with floats using template function: %ld\n",
         t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

// Third test: use templated structs with template function.

template<typename T>
struct vec3 {
  T x, y, z;
};

void test_int3() {
  vec3<int> a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = rand();
    a[i].y = rand();
    a[i].z = rand();
    b[i].x = rand();
    b[i].y = rand();
    b[i].z = rand();
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    vec3_sum(c[i], b[i], a[i]);
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with integers using template struct: %ld\n",
         t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

void test_float3() {
  vec3<float> a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = ((float)rand())/RAND_MAX;
    a[i].y = ((float)rand())/RAND_MAX;
    a[i].z = ((float)rand())/RAND_MAX;
    b[i].x = ((float)rand())/RAND_MAX;
    b[i].y = ((float)rand())/RAND_MAX;
    b[i].z = ((float)rand())/RAND_MAX;
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    vec3_sum(c[i], b[i], a[i]);
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with floats using template struct: %ld\n",
         t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

// Fourth test: use class/struct operator overload.

template<typename T>
struct vec3_with_op {
  T x, y, z;

  vec3_with_op<T> operator+(vec3_with_op<T> const &other) const {
    vec3_with_op<T> ret;
    ret.x = this->x + other.x;
    ret.y = this->y + other.y;
    ret.z = this->z + other.z;
  }
};

void test_int4() {
  vec3_with_op<int> a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = rand();
    a[i].y = rand();
    a[i].z = rand();
    b[i].x = rand();
    b[i].y = rand();
    b[i].z = rand();
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    c[i] = a[i] + b[i];
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with integers using operator overload: %ld\n",
         t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

void test_float4() {
  vec3_with_op<float> a[N_ELEM], b[N_ELEM], c[N_ELEM];
  unsigned int i;
  struct timeval t1, t2, t_diff;

  for (i = 0; i < N_ELEM; i++) {
    a[i].x = rand();
    a[i].y = rand();
    a[i].z = rand();
    b[i].x = rand();
    b[i].y = rand();
    b[i].z = rand();
  }

  gettimeofday(&t1, NULL);
  for (i = 0; i < N_ELEM; i++)
    c[i] = a[i] + b[i];
  gettimeofday(&t2, NULL);

  timersub(&t2, &t1, &t_diff);
  printf("Time with float using operator overload: %ld\n",
         t_diff.tv_sec*1000000 + t_diff.tv_usec);
}

#define STACK_SIZE 1000000000

int main() {
  struct rlimit rl;
  int res;

  srand(0);

  res = getrlimit(RLIMIT_STACK, &rl);
  if (res != 0) {
    printf("Error getting stack size! (%d)\n", res);
    exit(1);
  }

  rl.rlim_cur = STACK_SIZE;
  res = setrlimit(RLIMIT_STACK, &rl);
  if (res != 0) {
    printf("Error setting stack size! (%d)\n", res);
    exit(1);
  }

  printf("Test 1:\n");
  test_int1();
  test_float1();

  printf("\nTest 2:\n");
  test_int2();
  test_float2();

  printf("\nTest 3:\n");
  test_int3();
  test_float3();

  printf("\nTest 4:\n");
  test_int4();
  test_float4();
}
